#!/usr/bin/env python3
import aio_pika
import os
import sys
import traceback
import base64
import json
import asyncio
import socket
from . import MythicCommandBase
from . import PayloadBuilder
from pathlib import Path
from importlib import import_module, invalidate_caches
from .config import settings
from functools import partial

# set the global hostname variable
hostname = ""
output = ""
exchange = None
channel = None
queue = None
connection = None
operating_environment = "production"
container_files_path = ""

container_version = "12"
PyPi_version = "0.1.10"


def print_flush(msg, override: bool = False):
    global operating_environment
    if operating_environment in ["development", "testing", "staging"] or override:
        print(msg)
        sys.stdout.flush()


def import_all_agent_functions():
    import glob

    # Get file paths of all modules.
    modules = glob.glob("agent_functions/*.py")
    invalidate_caches()
    for x in modules:
        if not x.endswith("__init__.py") and x[-3:] == ".py":
            module = import_module("agent_functions." + Path(x).stem)
            for el in dir(module):
                if "__" not in el:
                    globals()[el] = getattr(module, el)


async def send_status(message="", command="", status="", username="", reference_id=""):
    global exchange
    # status is success or error
    try:
        message_body = aio_pika.Message(message.encode())
        # Sending the message
        await exchange.publish(
            message_body,
            routing_key="pt.status.{}.{}.{}.{}.{}.{}".format(
                hostname, command, reference_id, status, username, container_version
            ),
        )
    except Exception as e:
        print_flush("Exception in send_status: {}".format(str(traceback.format_exc())))
        sys.exit(1)


async def initialize_task(command_class: MythicCommandBase.CommandBase, 
    message_json: dict, 
    command: str, 
    reference_id: str, 
    username: str) -> MythicCommandBase.MythicTask:
    task = MythicCommandBase.MythicTask(
        message_json["task"],
        args=command_class.argument_class(command_line=message_json["params"],
            tasking_location=message_json["tasking_location"],
            raw_command_line=message_json["task"]["original_params"],
            task_dictionary=message_json["task"]),
    )
    try:
        # if tasking came from the command_line or an unknown source, call parse_arguments to deal with unknown text
        if task.args.tasking_location == "command_line":
            await task.args.parse_arguments()
        else:
            # tasking didn't come from command line, so if we have a special function to parse dictionary entries, use it
            if hasattr(task.args, 'parse_dictionary') and callable(task.args.parse_dictionary):
                # if we got tasking from a modal popup or from tab complete, then the task.args.command_line is a dictionary
                await task.args.parse_dictionary(json.loads(message_json['params']))
            else:
                # otherwise, we still just have to call the parse_arguments function
                # this way we don't break any existing command parsing
                await task.args.parse_arguments()
    except Exception as pa:
        message = {"task": task.to_json(), 
            "message": f"[-] {hostname} failed to parse arguments for {message_json['command']}: \n" + str(traceback.format_exc())}
        await send_status(
            message=json.dumps(message),
            command=command,
            status="parse_arguments_error",
            reference_id=reference_id,
            username=username
        )
        return None
    try:
        await task.args.verify_required_args_have_values()
    except Exception as va:
        message = {"task": task.to_json(), 
            "message": f"[-] {message_json['command']} has arguments with invalid values: \n" + str(traceback.format_exc())}
        await send_status(
            message=json.dumps(message),
            command=command,
            status="verify_arguments_error",
            reference_id=reference_id,
            username=username
        )
        return None
    task.parameter_group_name = task.args.get_parameter_group_name()
    return task


async def callback(message: aio_pika.IncomingMessage):
    global hostname
    global container_files_path
    async with message.process():
        # messages of the form: pt.task.PAYLOAD_TYPE.command
        pieces = message.routing_key.split(".")
        command = pieces[3]
        username = pieces[5]
        reference_id = pieces[4]
        if command == "create_payload_with_code":
            try:
                # pt.task.PAYLOAD_TYPE.create_payload_with_code.UUID
                message_json = json.loads(
                    message.body, strict=False
                )
                # go through all the data from rabbitmq to make the proper classes
                c2info_list = []
                for c2 in message_json["c2_profile_parameters"]:
                    params = c2.pop("parameters", None)
                    c2info_list.append(
                        PayloadBuilder.C2ProfileParameters(parameters=params, c2profile=c2)
                    )
                commands = PayloadBuilder.CommandList(message_json["commands"])
                for cls in PayloadBuilder.PayloadType.__subclasses__():
                    agent_builder = cls(
                        uuid=message_json["uuid"],
                        agent_code_path=Path(container_files_path),
                        c2info=c2info_list,
                        selected_os=message_json["selected_os"],
                        commands=commands,
                        wrapped_payload=message_json["wrapped_payload"],
                    )
                try:
                    await agent_builder.set_and_validate_build_parameters(
                        message_json["build_parameters"]
                    )
                    build_resp = await agent_builder.build()
                except Exception as b:
                    resp_message = {
                        "status": "error",
                        "build_message": "",
                        "build_stdout": "",
                        "build_stderr": "Error in agent creation: " + str(traceback.format_exc()),
                        "payload": "",
                    }
                    await send_status(
                        message=json.dumps(resp_message),
                        command="create_payload_with_code",
                        status="error",
                        reference_id=reference_id,
                        username=username,
                    )
                    return
                # we want to capture the build message as build_resp.get_build_message()
                # we also want to capture the final values the agent used for creating the payload, so collect them
                build_instances = agent_builder.get_build_instance_values()
                resp_message = {
                    "status": build_resp.get_status(),
                    "build_message": build_resp.get_build_message(),
                    "build_stdout": build_resp.get_build_stdout(),
                    "build_stderr": build_resp.get_build_stderr(),
                    "build_parameter_instances": build_instances,
                    "payload": base64.b64encode(build_resp.get_payload()).decode(
                        "utf-8"
                    ),
                }
                await send_status(
                    message=json.dumps(resp_message),
                    command="create_payload_with_code",
                    reference_id=reference_id,
                    status="success",
                    username=username,
                )

            except Exception as e:
                resp_message = {
                    "status": "error",
                    "build_message": "",
                    "build_stdout": "",
                    "build_stderr": str(traceback.format_exc()),
                    "payload": "",
                }
                await send_status(
                    message=json.dumps(resp_message),
                    command="create_payload_with_code",
                    status="error",
                    reference_id=reference_id,
                    username=username,
                )
        elif command == "command_transform":
            try:
                task = None
                # pt.task.PAYLOAD_TYPE.command_transform.taskID
                message_json = json.loads(
                    message.body, strict=False
                )
                final_task = None
                if "tasking_location" not in message_json:
                    message_json["tasking_location"] = "command_line"
                for cls in MythicCommandBase.CommandBase.__subclasses__():
                    if getattr(cls, "cmd") == message_json["command"]:
                        Command = cls(Path(container_files_path))
                        task = await initialize_task(command_class=Command, message_json=message_json, command=command, reference_id=reference_id, username=username)
                        if task is None:
                            return
                        if 'opsec_class' in dir(Command) and Command.opsec_class is not None and callable(Command.opsec_class.opsec_pre):
                            # the opsec_pre function is defined
                            if task.opsec_pre_blocked is None:
                                # this means we haven't run the function to see if it's been blocked or not
                                #   opsec_pre updates task itself
                                try:
                                    await Command.opsec_class().opsec_pre(task)
                                except Exception as ct:
                                    message = {"task": task.to_json()}
                                    if type(ct).__name__ == "Exception":
                                        # we're looking at a generic exception, probably one raised on purpose by the function
                                        message["message"] = f"[-] {hostname} ran into an error processing an opsec_pre check for {message_json['command']}: \n" + str(ct)
                                        await send_status(
                                            message=json.dumps(message),
                                            command="command_transform",
                                            status="opsec_pre_error",
                                            reference_id=reference_id,
                                            username=username
                                        )
                                    else:
                                        # we're probably looking at an actual error
                                        message["message"] = f"[-] {hostname} ran into an error processing and opsec_pre check for {message_json['command']}: \n" + str(ct) + "\n" + str(traceback.format_exc())
                                        await send_status(
                                            message=json.dumps(message),
                                            command="command_transform",
                                            status="opsec_pre_error",
                                            reference_id=reference_id,
                                            username=username
                                        )
                                    return
                                if task.opsec_pre_blocked is True:
                                    # the opsec_pre function decided to block the task
                                    if not task.opsec_pre_bypassed:
                                        # and not provide a bypass
                                        # send our current result and indicate we were blocked at opsec_pre
                                        message = {"task": task.to_json(), 
                                        "message": ""}
                                        await send_status(
                                            message=json.dumps(message),
                                            command="command_transform",
                                            status="opsec_pre_success",
                                            reference_id=reference_id,
                                            username=username
                                        )
                                        return
                                    # we threw up a block, but also set bypassed to true, so continue on
                            elif task.opsec_pre_blocked is True and not task.opsec_pre_bypassed:
                                # we already sent our opsec_pre data, we're blocked and not bypassed, just return
                                return
                        try:
                            final_task = await Command.create_tasking(task)
                        except Exception as ct:
                            message = {"task": task.to_json()}
                            if type(ct).__name__ == "Exception":
                                # we're looking at a generic exception, probably one raised on purpose by the function
                                message["message"] = f"[-] {hostname} ran into an error processing {message_json['command']}: \n" + str(ct)
                                await send_status(
                                    message=json.dumps(message),
                                    command="command_transform",
                                    status="create_tasking_error",
                                    reference_id=reference_id,
                                    username=username
                                )
                            else:
                                # we're probably looking at an actual error
                                message["message"] = f"[-] {hostname} ran into an error processing {message_json['command']}: \n" + str(ct) + "\n" + str(traceback.format_exc())
                                await send_status(
                                    message=json.dumps(message),
                                    command="command_transform",
                                    status="create_tasking_error",
                                    reference_id=reference_id,
                                    username=username
                                )
                            return
                        if "opsec_class" in dir(Command) and Command.opsec_class is not None and callable(Command.opsec_class.opsec_post):
                            # the opsec_post function is defined
                            if task.opsec_post_blocked is None:
                                # this means we haven't run the function to see if it's been blocked or not
                                #   opsec_post updates task itself
                                try:
                                    await Command.opsec_class().opsec_post(task)
                                except Exception as ct:
                                    message = {"task": task.to_json()}
                                    if type(ct).__name__ == "Exception":
                                        # we're looking at a generic exception, probably one raised on purpose by the function
                                        message["message"] = f"[-] {hostname} ran into an error processing an opsec_post check for {message_json['command']}: \n" + str(ct)
                                        await send_status(
                                            message=json.dumps(message),
                                            command="command_transform",
                                            status="opsec_post_error",
                                            reference_id=reference_id,
                                            username=username
                                        )
                                    else:
                                        # we're probably looking at an actual error
                                        message["message"] = f"[-] {hostname} ran into an error processing an opsec_post check for {message_json['command']}: \n" + str(ct) + "\n" + str(traceback.format_exc())
                                        await send_status(
                                            message=json.dumps(message),
                                            command="command_transform",
                                            status="opsec_post_error",
                                            reference_id=reference_id,
                                            username=username
                                        )
                                    return
                                if task.opsec_post_blocked and not task.opsec_post_bypassed:
                                    # send the results of the check
                                    message = {"task": task.to_json(), "message": ""}
                                    await send_status(
                                        message=json.dumps(message),
                                        command="command_transform",
                                        status="opsec_post_success",
                                        reference_id=reference_id,
                                        username=username
                                    )
                                    return
                        message = {"task": task.to_json(), "message": ""}
                        await send_status(
                            message=json.dumps(message),
                            command="command_transform",
                            status=final_task.status if final_task.status != "preprocessing" else "success",
                            reference_id=reference_id,
                            username=username
                        )
                        break
                if final_task is None:
                    message = {"task": task.to_json(),  "message": "Failed to find class where command_name = " + message_json["command"]}
                    await send_status(
                        message=json.dumps(message),
                        command="command_transform",
                        status="error_not_found",
                        reference_id=reference_id,
                        username=username
                    )
            except Exception as e:
                message = {"task": task.to_json() if task is not None else message_json["task"], "message": "[-] Mythic error while creating/running create_tasking: \n" + str(e)}
                await send_status(
                    message=json.dumps(message),
                    command="command_transform",
                    status="general_error",
                    reference_id=reference_id,
                    username=username
                )
                return
        elif command == "sync_classes":
            await sync_classes(reference_id)
            pass
        elif command == "process_container":
            try:
                # pt.task.PAYLOAD_TYPE.command_transform.taskID
                message_json = json.loads(message.body)
                final_task = None
                for cls in MythicCommandBase.CommandBase.__subclasses__():
                    if getattr(cls, "cmd") == message_json["command"]:
                        Command = cls(Path(container_files_path))
                        task = await initialize_task(command_class=Command, message_json=message_json, command=command, reference_id=reference_id, username=username)
                        if task is None:
                            return
                        agentResponse = MythicCommandBase.AgentResponse(task=task, response=message_json["response"])
                        await Command.process_response(agentResponse)
                        break
            except Exception as e:
                await send_status(
                    message="[-] Error while running process_response: \n" + str(e),
                    command="process_container",
                    status="error",
                    reference_id=reference_id,
                    username=username
                )
                return
        elif command == "exit_container":
            print_flush("[*] Got exit container command, exiting!", override=True)
            print_flush("[*] Container versions and information can be found at: https://docs.mythic-c2.net/customizing/payload-type-development/container-syncing#current-payloadtype-versions",override=True)
            print_flush("[*] If this is unexpected, check the Mythic UI for logs as to why this container is tasked to exit", override=True)
            sys.exit(0)
        elif command == "task_callback_function":
            try:
                # pt.task.PAYLOAD_TYPE.task_callback_function.taskID
                message_json = json.loads(
                    message.body, strict=False
                )
                task = None
                final_task = None
                for cls in MythicCommandBase.CommandBase.__subclasses__():
                    if getattr(cls, "cmd") == message_json["command"]:
                        Command = cls(Path(container_files_path))
                        task = await initialize_task(command_class=Command, message_json=message_json, command=command, reference_id=reference_id, username=username)
                        if task is None:
                            return
                        if hasattr(Command, message_json["function_name"]) and callable(getattr(Command, message_json["function_name"])):
                            # the message_json["function_name"] function is defined in the Command.callback_class and is callable
                            try:
                                final_task = await getattr(Command, message_json["function_name"])(task, message_json["subtask"], message_json["subtask_group_name"])
                            except Exception as fn:
                                message = {"task": task.to_json(),  
                                "updating_task": message_json["updating_task"],
                                "updating_piece": message_json["updating_piece"],
                                "message": f"[-] {hostname} hit an exception in callback function, {message_json['function_name']} for {message_json['command']}: \n" + str(traceback.format_exc())}
                                await send_status(
                                    message=json.dumps(message),
                                    command="task_callback_function",
                                    status="handler_error",
                                    reference_id=reference_id,
                                    username=username
                                )
                                return
                            # send the results of the check
                            message = {"task": final_task.to_json(),
                                "updating_task": message_json["updating_task"],
                                "updating_piece": message_json["updating_piece"],
                                "message": ""}
                            await send_status(
                                message=json.dumps(message),
                                command="task_callback_function",
                                status=final_task.status if final_task.status != "preprocessing" else "success",
                                reference_id=reference_id,
                                username=username
                            )
                        else:
                            message = {"task": task.to_json(), 
                            "updating_task": message_json["updating_task"],
                            "updating_piece": message_json["updating_piece"],
                            "message": f"[-] {hostname} failed to find callback function, {message_json['function_name']} for {message_json['command']}"}
                            await send_status(
                                message=json.dumps(message),
                                command="task_callback_function",
                                status="not_found_error",
                                reference_id=reference_id,
                                username=username
                            )
                            return
            except Exception as e:
                message = {"task": task.to_json() if task is not None else message_json["task"],
                    "message": "[-] Mythic error while creating/running task_callback_function: \n" + str(traceback.format_exc())}
                await send_status(
                    message=json.dumps(message),
                    command="task_callback_function",
                    status="generic_error",
                    reference_id=reference_id,
                    username=username
                )
                return
        else:
            print("Unknown command: {}".format(command))


async def sync_classes(reference_id: str = ""):
    try:
        commands = {}
        payload_type = {}
        print_flush("[*] About to import all of the agent's functions")
        import_all_agent_functions()
        print_flush("[*] Agent function import completed. Now to parse the Payload and Command information")
        for cls in PayloadBuilder.PayloadType.__subclasses__():
            payload_type = cls(agent_code_path=Path(container_files_path)).to_json()
            break
        for cls in MythicCommandBase.CommandBase.__subclasses__():
            commands[cls.cmd] = cls(Path(container_files_path)).to_json()
        payload_type["commands"] = commands
        print_flush("[*] Payload and Command parsing completed. Now sending sync to Mythic")
        await send_status(json.dumps(payload_type), command="sync_classes", status="success", username="", reference_id=reference_id)
    except Exception as e:
        print_flush("[-] Failed to sync classes, exiting container!\n" + str(traceback.format_exc()), override=True)
        await send_status(
            message="Error while syncing info: " + str(traceback.format_exc()),
            command="sync_classes",
            status="error",
            username="",
            reference_id=reference_id
        )
        sys.exit(1)


async def connect_to_rabbitmq():
    global hostname
    global exchange
    global channel
    global queue
    global connection
    global container_files_path
    if queue is not None:
        return
    hostname = settings.get("name", "hostname")
    if hostname == "hostname":
        hostname = socket.gethostname()
        print_flush("[*] Hostname specified as default 'hostname' value, thus will fetch and use the current hostname of the Docker container or computer\n", override=True)
    print_flush("[*] Setting hostname (which should match payload type name exactly) to: " + hostname, override=True)
    container_files_path = os.path.abspath(settings.get("container_files_path", "/Mythic/"))
    if not os.path.exists(container_files_path):
        print_flush(f"[-] Specified mythic_container_files_path, {container_files_path},  does not exist", override=True)
        print_flush(f"[-] If this isn't in a Docker container, then this should be the path to your PayloadType folder", override=True)
        print_flush(f"[-] ex: C:\\Users\\username\\Desktop\\AgentName\\", override=True)
        print_flush(f"[-] ex: /Users/username/AgentName/", override=True)
        sys.exit(1)
    while True:
        try:
            print_flush("[*] Trying to connect to rabbitmq at: " + settings.get("host", "127.0.0.1") + ":" + str(settings.get("port", 5672)),  override=True)
            connection = await aio_pika.connect_robust(
                host=settings.get("host", "127.0.0.1"),
                port=settings.get("port", 5672),
                login=settings.get("username", "mythic_user"),
                password=settings.get("password", "mythic_password"),
                virtualhost=settings.get("virtual_host", "mythic_vhost"),
            )
            print_flush("[*] connecting to channel, then declaring exchange")
            channel = await connection.channel()
            # declare our heartbeat exchange that everybody will publish to, but only the mythic server will are about
            print_flush("[*] declaring exchange, then declaring queue")
            exchange = await channel.declare_exchange(
                "mythic_traffic", aio_pika.ExchangeType.TOPIC
            )
            # get a random queue that only the mythic server will use to listen on to catch all heartbeats
            queue = await channel.declare_queue(hostname + "_tasking")
            print_flush("[*] binding to exchange, then ready to go")
            # bind the queue to the exchange so we can actually catch messages
            await queue.bind(
                exchange="mythic_traffic", routing_key="pt.task.{}.#".format(hostname)
            )

            # just want to handle one message at a time so we can clean up and be ready
            await channel.set_qos(prefetch_count=int(settings.get("queue_prefetch", "10")))
            print_flush("[+] Ready to go!",  override=True)
            return
        except Exception as e:
            print_flush(str(e))
            await asyncio.sleep(2)
            continue


async def heartbeat():
    global exchange
    global queue
    global channel
    global hostname
    try:
        if exchange is not None:
            while True:
                try:
                    queue = await channel.get_queue(hostname + "_tasking")
                    await exchange.publish(
                        aio_pika.Message("".encode()),
                        routing_key="pt.heartbeat.{}.{}.{}".format(hostname, container_version, str(queue.declaration_result.consumer_count)),
                    )
                    await asyncio.sleep(10)
                except Exception as e:
                    print_flush("[*] heartbeat - exception in heartbeat message loop:\n " + str(traceback.format_exc()), override=True)
                    sys.exit(1)
        else:
            print_flush("[-] Failed to process heartbeat functionality, exiting container:\n " + str(traceback.format_exc()), override=True)
            sys.exit(1)
    except Exception as h:
        print_flush("[-] Failed to process heartbeat functionality, exiting container:\n " + str(traceback.format_exc()), override=True)
        sys.exit(1)


async def mythic_service():
    global hostname
    global queue

    try:
        print_flush("[*] mythic_service - Waiting for messages in mythic_service with version {}.".format(container_version), override=True)
        task = queue.consume(callback)
        print_flush("[*] mythic_service - total instances of {} container running: {}".format(hostname, queue.declaration_result.consumer_count + 1), override=True)
        print_flush("[*] Now parsing agent_functions python files to sync with Mythic server")
        await sync_classes()
        print_flush("[*] Successfuly synced with Mythic server, waiting for tasking")
        result = await asyncio.wait_for(task, None)
    except Exception as e:
        print_flush("[-] mythic_service - exception, exiting container\n " + str(traceback.format_exc()))
        sys.exit(1)


async def rabbit_mythic_rpc_callback(
    exchange: aio_pika.Exchange, message: aio_pika.IncomingMessage
):
    async with message.process():
        try:
            message_json = json.loads(message.body.decode())
            # request = { "action": "function_name", "command": "command name"}
            response = {"status": "error", "error": "command's dynamic_query_function not found or not callable"}
            for cls in MythicCommandBase.CommandBase.__subclasses__():
                if getattr(cls, "cmd") == message_json["command"]:
                    Command = cls(Path(container_files_path))
                    CommandArgs = Command.argument_class("")
                    # now iterate over the args to find the one with name=request["action"]
                    for command_param in CommandArgs.args:
                        if command_param.name == message_json["action"] and callable(command_param.dynamic_query_function):
                            response = await command_param.dynamic_query_function(message_json["callback"])
                    break
        except Exception as e:
            print_flush("[-] Exception trying to respond to dynamic_query_function:\n" + str(traceback.format_exc()), override=True)
            response = {"status": "error", "error": str(e)}
        try:
            await exchange.publish(
                aio_pika.Message(body=json.dumps(response).encode(), correlation_id=message.correlation_id),
                routing_key=message.reply_to,
            )
        except Exception as e:
            print_flush(
                "[-] Exception trying to send message back to container for rpc! " + str(sys.exc_info()[-1].tb_lineno) + " " + str(e), override=True
            )

async def connect_and_consume_mythic_rpc():
    global channel
    global hostname
    # get a random queue that only the apfell server will use to listen on to catch all heartbeats
    print_flush("[*] Declaring container specific rpc queue in connect_and_consume_mythic_rpc")
    queue = await channel.declare_queue("{}_mythic_rpc_queue".format(hostname), auto_delete=True)
    await channel.set_qos(prefetch_count=int(settings.get("queue_prefetch_rpc", "10")))
    try:
        print_flush("[*] Starting to consume messages in connect_and_consume_mythic_rpc")
        task = await queue.consume(
            partial(rabbit_mythic_rpc_callback, channel.default_exchange)
        )
    except Exception as e:
        print_flush("[-] Exception in connect_and_consume_mythic_rpc .consume: {}".format(str(sys.exc_info()[-1].tb_lineno) + " " + str(e)), override=True)
        sys.exit(1)

# start our service
def start_service_and_heartbeat(**kwargs):
    global operating_environment
    if "debug" in kwargs:
        print_flush("[*] the `debug` argument is deprecated, instead set the `MYTHIC_ENVIRONMENT` variable directly for the container or for Mythic/.env", override=True)
    operating_environment = settings.get("environment", "production")
    if "debug" in kwargs and kwargs["debug"]:
        operating_environment = "testing"
        print_flush("[*] Setting MYTHIC_ENVIRONMENT=Testing based on deprecated debug=True argument", override=True)
    if operating_environment == "production":
        print_flush("[*] To enable debug logging, set `MYTHIC_ENVIRONMENT` variable to `testing`", override=True)
    get_version_info()
    loop = asyncio.get_event_loop()
    connect_task = loop.create_task(connect_to_rabbitmq())
    loop.run_until_complete(connect_task)
    asyncio.gather(heartbeat(), mythic_service(), connect_and_consume_mythic_rpc())
    loop.run_forever()

def get_version_info():
    print_flush("[*] Mythic PayloadType Version: " + container_version, override=True)
    print_flush("[*] PayloadType PyPi Version: " + PyPi_version, override=True)



from abc import abstractmethod, ABCMeta
import json
from enum import Enum
import base64
import uuid
from pathlib import Path
from .PayloadBuilder import SupportedOS


class MythicStatus():
    Error = "error"
    Completed = "completed"
    Processed = "processed"
    Processing = "processing"
    Preprocessing = "preprocessing"
    Delegating = "delegating subtasks"
    CallbackError = "task callback error"
    Success = "success"
    def __init__(self, status: str):
        self.status = status
    def __str__(self):
        return self.status
    def __eq__(self, obj):
        # check if self.status == obj
        if isinstance(obj, str):
            return self.status == obj 
        elif isinstance(obj, MythicStatus):
            return self.status == obj.status 
        elif isinstance(obj, MythicRPCStatus):
            return self.status == obj.status
        else:
            return False

class MythicRPCStatus():
    Success = "success"
    Error = "error"
    def __init__(self, status: str):
        self.status = status
    def __str__(self):
        return self.status
    def __eq__(self, obj):
        if isinstance(obj, str):
            return self.status == obj 
        elif isinstance(obj, MythicRPCStatus):
            return self.status == obj.status 
        elif ininstance(obj, MythicStatus):
            return self.status == obj.status
        else:
            return False

class ParameterType(str, Enum):
    String = "String"
    Boolean = "Boolean"
    File = "File"
    Array = "Array"
    ChooseOne = "Choice"
    ChooseMultiple = "ChoiceMultiple"
    Credential_JSON = "Credential-JSON"
    Credential_Account = "Credential-Account"
    Credential_Realm = "Credential-Realm"
    Credential_Type = "Credential-Type"
    Credential_Value = "Credential-Credential"
    Number = "Number"
    Payload = "PayloadList"
    ConnectionInfo = "AgentConnect"
    LinkInfo = "LinkInfo"

class CommandAttributes():
    def __init__(self,
        spawn_and_injectable: bool = False,
        supported_os: [SupportedOS] = None,
        builtin: bool = False,
        suggested_command: bool = False,
        load_only: bool = False,
        filter_by_build_parameter: dict = {},
        **kwargs):
        self.spawn_and_injectable = spawn_and_injectable
        self.supported_os = supported_os
        self.builtin = builtin
        self.suggested_command = suggested_command
        self.load_only = load_only
        self.filter_by_build_parameter = filter_by_build_parameter
        self.additional_items = {}
        for k,v in kwargs.items():
            self.additional_items[k] = v
    
    def to_json(self):
        r = {}
        if self.spawn_and_injectable is not None:
            r["spawn_and_injectable"] = self.spawn_and_injectable
        else:
            r["spawn_and_injectable"] = False
        if self.supported_os is not None:
            r["supported_os"] = [str(x) for x in self.supported_os]
        else:
            r["supported_os"] = []
        r["builtin"] = self.builtin
        r["suggested_command"] = self.suggested_command
        r["load_only"] = self.load_only
        r["filter_by_build_parameter"] = self.filter_by_build_parameter
        r = {**r, **self.additional_items}
        return r

class ParameterGroupInfo:
    def __init__(
        self,
        required: bool = True,
        group_name: str = "Default",
        ui_position: int = None,
        **kwargs
    ):  
        self.required = required 
        self.group_name = group_name 
        self.ui_position = ui_position
        self.additional_info = {}
        for k,v in kwargs.items():
            self.additional_info[k] = v

    def to_json(self):
        r = {}
        r["required"] = self.required 
        r["group_name"] = self.group_name 
        r["ui_position"] = self.ui_position
        r = {**r, **self.additional_info}
        return r

class CommandParameter:
    def __init__(
        self,
        name: str,
        type: ParameterType,
        display_name: str = None,
        cli_name: str = None,
        description: str = "",
        choices: [any] = None,
        default_value: any = None,
        validation_func: callable = None,
        value: any = None,
        supported_agents: [str] = None,
        supported_agent_build_parameters: dict = None,
        choice_filter_by_command_attributes: dict = None,
        choices_are_all_commands: bool = False,
        choices_are_loaded_commands: bool = False,
        dynamic_query_function: callable = None,
        parameter_group_info: [ParameterGroupInfo] = None
    ):
        self.name = name
        if display_name is None:
            self.display_name = name
        else:
            self.display_name = display_name
        if cli_name is None:
            self.cli_name = name 
        else:
            self.cli_name = cli_name
        self.type = type
        self.user_supplied = False # keep track of if this is using the default value or not
        self.description = description
        if choices is None:
            self.choices = []
        else:
            self.choices = choices
        self.validation_func = validation_func
        if value is None:
            self._value = default_value
        else:
            self.value = value
        self.default_value = default_value
        self.supported_agents = supported_agents if supported_agents is not None else []
        self.supported_agent_build_parameters = supported_agent_build_parameters if supported_agent_build_parameters is not None else {}
        self.choice_filter_by_command_attributes = choice_filter_by_command_attributes if choice_filter_by_command_attributes is not None else {}
        self.choices_are_all_commands = choices_are_all_commands
        self.choices_are_loaded_commands = choices_are_loaded_commands
        self.dynamic_query_function = dynamic_query_function
        if not callable(dynamic_query_function) and dynamic_query_function is not None:
            raise Exception("dynamic_query_function is not callable")
        self.parameter_group_info = parameter_group_info
        if self.parameter_group_info is None:
            self.parameter_group_info = [ParameterGroupInfo()]

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, type):
        self._type = type

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, description):
        self._description = description

    @property
    def choices(self):
        return self._choices

    @choices.setter
    def choices(self, choices):
        self._choices = choices

    @property
    def validation_func(self):
        return self._validation_func

    @validation_func.setter
    def validation_func(self, validation_func):
        self._validation_func = validation_func

    @property
    def supported_agents(self):
        return self._supported_agents

    @supported_agents.setter
    def supported_agents(self, supported_agents):
        self._supported_agents = supported_agents
        
    @property
    def supported_agent_build_parameters(self):
        return self._supported_agent_build_parameters

    @supported_agent_build_parameters.setter
    def supported_agent_build_parameters(self, supported_agent_build_parameters):
        self._supported_agent_build_parameters = supported_agent_build_parameters

    @property
    def dynamic_query_func(self):
        return self._dynamic_query_func

    @dynamic_query_func.setter
    def dynamic_query_func(self, dynamic_query_func):
        self._dynamic_query_func = dynamic_query_func

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if value is not None:
            type_validated = TypeValidators().validate(self.type, value)
            if self.validation_func is not None:
                try:
                    self.validation_func(type_validated)
                    self._value = type_validated
                except Exception as e:
                    raise ValueError(
                        "Failed validation check for parameter {} with value {}".format(
                            self.name, str(value)
                        )
                    )
                self.user_supplied = True
                return
            else:
                # now we do some verification ourselves based on the type
                self._value = type_validated
                self.user_supplied = True
                return
        self._value = value
        self.user_supplied = True 
        return

    @property
    def parameter_group_info(self):
        return self._parameter_group_info

    @parameter_group_info.setter
    def parameter_group_info(self, parameter_group_info):
        self._parameter_group_info = parameter_group_info
    

    def to_json(self):
        return {
            "name": self._name,
            "display_name": self.display_name,
            "cli_name": self.cli_name.replace(" ", "-"),
            "type": self._type.value,
            "description": self._description,
            "choices": "\n".join(self._choices),
            "default_value": self.default_value,
            "value": self.value,
            "supported_agents": ",".join(self._supported_agents),
            "supported_agent_build_parameters": self._supported_agent_build_parameters,
            "choices_are_loaded_commands": self.choices_are_loaded_commands,
            "choices_are_all_commands": self.choices_are_all_commands,
            "choice_filter_by_command_attributes": self.choice_filter_by_command_attributes,
            "dynamic_query_function": self.dynamic_query_function.__name__ if callable(self.dynamic_query_function) else None,
            "parameter_group_info": [x.to_json() for x in self.parameter_group_info] if self.parameter_group_info is not None else [ParameterGroupInfo().to_json()]
        }

class TypeValidators:
    def validateString(self, val):
        return str(val)

    def validateNumber(self, val):
        try:
            return int(val)
        except:
            return float(val)

    def validateBoolean(self, val):
        if isinstance(val, bool):
            return val
        else:
            raise ValueError("Value isn't bool")

    def validateFile(self, val):
        try:  # check if the file is actually a file-id
            uuid_obj = uuid.UUID(val, version=4)
            return str(uuid_obj)
        except ValueError:
            pass
        return base64.b64decode(val)

    def validateArray(self, val):
        if isinstance(val, list):
            return val
        else:
            raise ValueError("value isn't array")

    def validateCredentialJSON(self, val):
        if isinstance(val, dict):
            return val
        else:
            raise ValueError("value ins't a dictionary")

    def validatePass(self, val):
        return val

    def validateChooseMultiple(self, val):
        if isinstance(val, list):
            return val
        else:
            raise ValueError("Choices aren't in a list")

    def validatePayloadList(self, val):
        return str(uuid.UUID(val, version=4))

    def validateAgentConnect(self, val):
        if isinstance(val, dict):
            return val
        else:
            raise ValueError("Not instance of dictionary")

    switch = {
        "String": validateString,
        "Number": validateNumber,
        "Boolean": validateBoolean,
        "File": validateFile,
        "Array": validateArray,
        "Credential-JSON": validateCredentialJSON,
        "Credential-Account": validatePass,
        "Credential-Realm": validatePass,
        "Credential-Type": validatePass,
        "Credential-Credential": validatePass,
        "Choice": validatePass,
        "ChoiceMultiple": validateChooseMultiple,
        "PayloadList": validatePayloadList,
        "AgentConnect": validateAgentConnect,
        "LinkInfo": validateAgentConnect
    }

    def validate(self, type: ParameterType, val: any):
        return self.switch[type.value](self, val)

class TaskArguments(metaclass=ABCMeta):
    def __init__(self, 
        command_line: str, 
        tasking_location: str = "command_line", 
        raw_command_line: str = "",
        task_dictionary: dict = {}):
        self.command_line = str(command_line)
        self.tasking_location = tasking_location
        self.raw_command_line = raw_command_line
        self.task_dictionary = task_dictionary
        self.parameter_group_name = None

    @property
    def args(self):
        return self._args

    @args.setter
    def args(self, args):
        self._args = args

    def get_arg(self, key: str):
        for arg in self.args:
            if arg.name == key:
                return arg.value
        return None

    def has_arg(self, key: str) -> bool:
        for arg in self.args:
            if arg.name == key:
                return True
        return False

    def get_command_line(self) -> str:
        return self.command_line

    def get_raw_command_line(self) -> str:
        return self.raw_command_line

    def get_tasking_location(self) -> str:
        return self.tasking_location

    def is_empty(self) -> bool:
        return len(self.args) == 0

    def add_arg(self, key: str, value, type: ParameterType = None, parameter_group_info: [ParameterGroupInfo] = [ParameterGroupInfo()]):
        found = False
        for arg in self.args:
            if arg.name == key:
                arg.value = value 
                if type is not None:
                    arg.type = type 
                found = True
        if not found:
            if type is not None:
                self.args.append(CommandParameter(name=key, type=type, value=value, parameter_group_info=parameter_group_info))
            else:
                self.args.append(CommandParameter(name=key, type=ParameterType.String, value=value, parameter_group_info=parameter_group_info))

    def set_arg(self, key: str, value):
        found = False 
        for arg in self.args:
            if arg.name == key:
                arg.value = value 
                found = True 
        if not found:
            self.add_arg(key, value)

    def rename_arg(self, old_key: str, new_key: str):
        for arg in self.args:
            if arg.name == old_key:
                arg.name = new_key
                return
        raise Exception("{} not a valid parameter name".format(old_key))

    def remove_arg(self, key: str):
        self.args = [x for x in self.args if x.name != key]

    def to_json(self):
        return [x.to_json() for x in self.args]

    def load_args_from_json_string(self, command_line: str) -> None:
        temp_dict = json.loads(command_line)
        for k, v in temp_dict.items():
            for arg in self.args:
                if arg.name == k or arg.cli_name == k:
                    arg.value = v

    def load_args_from_dictionary(self, dictionary) -> None:
        for k, v in dictionary.items():
            for arg in self.args:
                if arg.name == k or arg.cli_name == k:
                    arg.value = v

    def get_parameter_group_name(self) -> str:
        if self.parameter_group_name is not None:
            return self.parameter_group_name
        groupNameOptions = []
        suppliedArgNames = []
        if len(self.args) == 0:
            return "Default"
        for arg in self.args:
            for group_info in arg.parameter_group_info:
                if group_info.group_name not in groupNameOptions:
                    groupNameOptions.append(group_info.group_name)
        for arg in self.args:
            # when determining the group we're in, only look at arguments that have values that were set by the user
            # default values don't count
            if arg.value is not None and arg.user_supplied:
                suppliedArgNames.append(arg.name)
                groupNameIntersection = []
                for group_info in arg.parameter_group_info:
                    if group_info.group_name in groupNameOptions:
                        groupNameIntersection.append(group_info.group_name)
                groupNameOptions = groupNameIntersection
            # this gives us the groups that our currently supplied commands belong to, but doesn't account for some groups still needing other parameters
            # need to loop through available options and see if we have all of the needed parameters
        if len(groupNameOptions) == 0:
            raise ValueError(f"Supplied Arguments, {suppliedArgNames}, don't match any parameter group") 
        elif len(groupNameOptions) > 1:
            finalMatchingGroupNames = []
            for groupNameOption in groupNameOptions:
                has_all_values = True
                for arg in self.args:
                    for group_info in arg.parameter_group_info:
                        if group_info.group_name == groupNameOption:
                            if group_info.required and not arg.user_supplied:
                                # one of our matching parameter groups that we've supplied values for requires this parameter, but the user didn't supply one
                                # so this parameter group can't be the one we use
                                has_all_values = False 
                if has_all_values:
                    finalMatchingGroupNames.append(groupNameOption)
            if len(finalMatchingGroupNames) == 0:
                raise ValueError(f"Supplied Arguments, {suppliedArgNames}, match more than one parameter group, {groupNameOptions}, and all require at least one more value from the user")
            elif len(finalMatchingGroupNames) > 1:
                raise ValueError(f"Supplied Arguments, {suppliedArgNames}, match more than one parameter group, {finalMatchingGroupNames}")
            else:
                return finalMatchingGroupNames[0]
            raise ValueError(f"Supplied Arguments, {suppliedArgNames}, match more than one parameter group: {groupNameOptions}")
        else:
            return groupNameOptions[0]

    def get_parameter_group_arguments(self) -> [CommandParameter]:
        groupName = self.get_parameter_group_name()
        group_arguments = []
        for arg in self.args:
            matched_arg = False
            for group_info in arg.parameter_group_info:
                if group_info.group_name == groupName:
                    matched_arg = True 
            if matched_arg:
                group_arguments.append(arg)
        return group_arguments

    async def verify_required_args_have_values(self) -> bool:
        # first we have to establish which parameter group we're in
        groupName = self.get_parameter_group_name()
        for arg in self.args:
            matched_arg = False
            arg_required = False
            for group_info in arg.parameter_group_info:
                if group_info.group_name == groupName:
                    matched_arg = True 
                    arg_required = group_info.required 
            if matched_arg:
                if arg.value is None:
                    arg.value = arg.default_value
                if arg_required and arg.value is None:
                    raise ValueError("Required arg {} has no value".format(arg.name))
        return True

    def __str__(self) -> str:
        if len(self.args) > 0:
            try:
                groupName = self.get_parameter_group_name()
            except Exception as e:
                return self.command_line
            temp = {}
            for arg in self.args:
                matched_arg = False
                for group_info in arg.parameter_group_info:
                    if group_info.group_name == groupName:
                        matched_arg = True 
                if matched_arg:
                    if isinstance(arg.value, bytes):
                        temp[arg.name] = base64.b64encode(arg.value).decode()
                    else:
                        temp[arg.name] = arg.value
            return json.dumps(temp)
        else:
            return self.command_line

    @abstractmethod
    async def parse_arguments(self) -> None:
        pass

class Callback:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

class BrowserScript:
    # if a browserscript is specified as part of a PayloadType, then it's a support script
    # if a browserscript is specified as part of a command, then it's for that command
    def __init__(self, script_name: str, author: str = None, for_new_ui: bool = False):
        self.script_name = script_name
        self.author = author
        self.for_new_ui = for_new_ui

    def to_json(self, base_path: Path):
        try:
            code_file = (
                base_path
                / "mythic"
                / "browser_scripts"
                / "{}.js".format(self.script_name)
            )
            if code_file.exists():
                code = code_file.read_bytes()
                code = base64.b64encode(code).decode()
            else:
                raise Exception("Code for Browser Script, " + self.script_name + ", does not exist on disk at path: " + str(code_file))
                code = ""
            return {"script": code, "name": self.script_name, "author": self.author, "for_new_ui": self.for_new_ui}
        except Exception as e:
            return {"script": str(e), "name": self.script_name, "author": self.author, "for_new_ui": self.for_new_ui}

class MythicTask:
    def __init__(
        self, taskinfo: dict, args: TaskArguments
    ):
        self.id = taskinfo["id"]
        self.original_params = taskinfo["original_params"]
        self.completed = taskinfo["completed"]
        self.callback = Callback(**taskinfo["callback"])
        self.agent_task_id = taskinfo["agent_task_id"]
        self.token = taskinfo["token"]
        self.operator = taskinfo["operator"]
        self.opsec_pre_blocked = taskinfo["opsec_pre_blocked"]
        self.opsec_pre_message = taskinfo["opsec_pre_message"]
        self.opsec_pre_bypassed = taskinfo["opsec_pre_bypassed"]
        self.opsec_pre_bypass_role = taskinfo["opsec_pre_bypass_role"]
        self.opsec_pre_bypass_user = taskinfo["opsec_pre_bypass_user"]
        self.opsec_post_blocked = taskinfo["opsec_post_blocked"]
        self.opsec_post_message = taskinfo["opsec_post_message"]
        self.opsec_post_bypassed = taskinfo["opsec_post_bypassed"]
        self.opsec_post_bypass_role = taskinfo["opsec_post_bypass_role"]
        self.opsec_post_bypass_user = taskinfo["opsec_post_bypass_user"]
        self.display_params = taskinfo["display_params"]
        self.command_name = taskinfo["command_name"]
        self.args = args
        self.manual_args = None
        self.status = MythicStatus.Preprocessing
        if 'status'in taskinfo and taskinfo['status'] is not None:
            self.status = MythicStatus(taskinfo['status'])
        self.tasking_location = taskinfo["tasking_location"] if "tasking_location" in taskinfo else "command_line"
        self.stdout = taskinfo["stdout"] if "stdout" in taskinfo else ""
        self.stderr = taskinfo["stderr"] if "stderr" in taskinfo else ""
        self.subtask_callback_function = taskinfo["subtask_callback_function"]
        self.group_callback_function = taskinfo["group_callback_function"]
        self.completed_callback_function = taskinfo["completed_callback_function"]
        self.subtask_group_name = taskinfo["subtask_group_name"]
        self.parameter_group_name = taskinfo["parameter_group_name"]
        # self.tags is an array of tags to associate with the task
        self.tags = taskinfo["tags"]

    def get_status(self) -> MythicStatus:
        return self.status

    def set_status(self, status: MythicStatus):
        self.status = status
        
    def set_stdout(self, stdout: str):
        self.stdout = stdout
    
    def set_stderr(self, stderr: str):
        self.stderr = stderr

    # if you call override_args with your own values, then we won't use the standard JSON string from self.args
    #   this combined with command_name can allow you to completely set what gets sent to your agent
    def override_args(self, args: str):
        self.manual_args = args

    def __str__(self):
        return json.dumps(self.to_json())

    def to_json(self):
        subtask_callback_function = self.subtask_callback_function
        if callable(subtask_callback_function):
            subtask_callback_function = subtask_callback_function.__name__
        group_callback_function = self.group_callback_function
        if callable(group_callback_function):
            group_callback_function = group_callback_function.__name__
        completed_callback_function = self.completed_callback_function
        if callable(completed_callback_function):
            completed_callback_function = completed_callback_function.__name__
        command_args = str(self.args)
        if self.manual_args is not None:
            command_args = self.manual_args
        return {
            "args": command_args,
            "stdout": self.stdout,
            "stderr": self.stderr,
            "opsec_pre_blocked": self.opsec_pre_blocked,
            "opsec_pre_message": self.opsec_pre_message,
            "opsec_pre_bypass_role": self.opsec_pre_bypass_role,
            "opsec_pre_bypassed": self.opsec_pre_bypassed,
            "opsec_post_blocked": self.opsec_post_blocked,
            "opsec_post_message": self.opsec_post_message,
            "opsec_post_bypass_role": self.opsec_post_bypass_role,
            "opsec_post_bypassed": self.opsec_post_bypassed,
            "display_params": self.display_params,
            "subtask_callback_function": subtask_callback_function,
            "group_callback_function": group_callback_function,
            "completed_callback_function": completed_callback_function,
            "subtask_group_name": self.subtask_group_name,
            "command_name": self.command_name,
            "parameter_group_name": self.parameter_group_name,
            "tags": self.tags
                          }

class AgentResponse:
    def __init__(self, response: any, task: MythicTask):
        self.response = response
        self.task = task

class CommandOPSEC(metaclass=ABCMeta):
    def __init__(self, 
        injection_method: str = "",
        process_creation: str = "",
        authentication: str = "",
        **kwargs):
        self.injection_method = injection_method
        self.process_creation = process_creation
        self.authentication = authentication
        self.additional_items = {}
        for k,v in kwargs.items():
            self.additional_items[k] = v

    @property
    @abstractmethod
    def injection_method(self):
        pass

    @property
    @abstractmethod
    def process_creation(self):
        pass

    @property
    @abstractmethod
    def authentication(self):
        pass

    def to_json(self):
        temp = {}
        temp["injection_method"] = self.injection_method
        temp["process_creation"] = self.process_creation
        temp["authentication"] = self.authentication
        temp = {**temp, **self.additional_items}
        return temp

    @abstractmethod
    async def opsec_pre(self, task: MythicTask):
        pass

    @abstractmethod
    async def opsec_post(self, task: MythicTask):
        pass

class CommandBase(metaclass=ABCMeta):
    def __init__(self, agent_code_path: Path):
        self.base_path = agent_code_path
        self.agent_code_path = agent_code_path / "agent_code"

    @property
    @abstractmethod
    def cmd(self):
        pass

    @property
    @abstractmethod
    def needs_admin(self):
        pass

    @property
    @abstractmethod
    def help_cmd(self):
        pass

    @property
    @abstractmethod
    def description(self):
        pass

    @property
    @abstractmethod
    def version(self):
        pass

    @property
    def supported_ui_features(self):
        pass

    @property
    @abstractmethod
    def author(self):
        pass

    @property
    @abstractmethod
    def argument_class(self):
        pass

    @property
    @abstractmethod
    def attackmapping(self):
        pass

    @property
    def browser_script(self):
        pass
    
    @property
    def attributes(self):
        pass
    
    @property
    def opsec_class(self):
        pass

    @property
    def script_only(self):
        pass

    @abstractmethod
    async def create_tasking(self, task: MythicTask) -> MythicTask:
        pass

    @abstractmethod
    async def process_response(self, response: AgentResponse):
        pass

    def to_json(self):
        params = self.argument_class("").to_json()
        opsec = self.opsec_class().to_json() if self.opsec_class is not None else {}
        if self.browser_script is not None:
            if isinstance(self.browser_script, list):
                bscript = {"browser_script": [x.to_json(self.base_path) for x in self.browser_script]}
            else:
                bscript = {"browser_script": [self.browser_script.to_json(self.base_path)]}
        else:
            bscript = {"browser_script": []}
        if self.attributes is None:
            attributes = CommandAttributes()
        else:
            attributes = self.attributes
        return {
            "cmd": self.cmd,
            "needs_admin": self.needs_admin,
            "help_cmd": self.help_cmd,
            "description": self.description,
            "version": self.version,
            "supported_ui_features": self.supported_ui_features if self.supported_ui_features is not None else [],
            "author": self.author,
            "attack": [{"t_num": a} for a in self.attackmapping],
            "parameters": params,
            "opsec": opsec,
            "attributes": attributes.to_json(),
            "script_only": self.script_only if self.script_only is not None else False,
            **bscript,
        }



from enum import Enum
from abc import abstractmethod
from pathlib import Path
import base64
from . import MythicCommandBase


class BuildStatus():
    Success = "success"
    Error = "error"

class SupportedOS():
    Windows = "Windows"
    MacOS = "macOS"
    Linux = "Linux"
    WebShell = "WebShell"
    Chrome = "Chrome"

    def __init__(self, os: str):
        self.os = os
    def __str__(self):
        return self.os


class BuildParameterType(str, Enum):
    String = "String"
    ChooseOne = "ChooseOne"
    Boolean = "Boolean"
    ChooseMultiple = "ChoiceMultiple"


class BuildParameter:
    def __init__(
        self,
        name: str,
        parameter_type: BuildParameterType = None,
        description: str = None,
        required: bool = None,
        verifier_regex: str = None,
        default_value: any = None,
        choices: [str] = None,
        value: any = None,
        verifier_func: callable = None,
    ):
        self.name = name
        self.verifier_func = verifier_func
        self.parameter_type = (
            parameter_type if parameter_type is not None else MythicCommandBase.ParameterType.String
        )
        self.description = description if description is not None else ""
        self.required = required if required is not None else True
        self.verifier_regex = verifier_regex if verifier_regex is not None else ""
        self.default_value = default_value
        if value is None:
            self.value = default_value
        else:
            self.value = value
        self.choices = choices

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def parameter_type(self):
        return self._parameter_type

    @parameter_type.setter
    def parameter_type(self, parameter_type):
        self._parameter_type = parameter_type

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, description):
        self._description = description

    @property
    def required(self):
        return self._required

    @required.setter
    def required(self, required):
        self._required = required

    @property
    def verifier_regex(self):
        return self._verifier_regex

    @verifier_regex.setter
    def verifier_regex(self, verifier_regex):
        self._verifier_regex = verifier_regex

    @property
    def default_value(self):
        return self._default_value

    @default_value.setter
    def default_value(self, default_value):
        self._default_value = default_value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if value is None:
            self._value = value
        else:
            if self.verifier_func is not None:
                self.verifier_func(value)
                self._value = value
            else:
                if isinstance(value, str) and value.lower() == "false":
                    self._value = False
                elif isinstance(value, str) and value.lower() == "true":
                    self._value = True
                else:
                    self._value = value

    def to_json(self):
        return {
            "name": self._name,
            "parameter_type": self._parameter_type.value,
            "description": self._description,
            "required": self._required,
            "verifier_regex": self._verifier_regex,
            "parameter": self._default_value
            if self._parameter_type in [BuildParameterType.String, BuildParameterType.Boolean]
            else "\n".join(self.choices),
        }


class C2ProfileParameters:
    def __init__(self, c2profile: dict, parameters: dict = None):
        self.parameters = {}
        self.c2profile = c2profile
        if parameters is not None:
            self.parameters = parameters

    def get_parameters_dict(self):
        return self.parameters

    def get_c2profile(self):
        return self.c2profile


class CommandList:
    def __init__(self, commands: [str] = None):
        self.commands = []
        if commands is not None:
            self.commands = commands

    def get_commands(self) -> [str]:
        return self.commands

    def remove_command(self, command: str):
        self.commands.remove(command)

    def add_command(self, command: str):
        for c in self.commands:
            if c == command:
                return
        self.commands.append(command)

    def clear(self):
        self.commands = []


class BuildResponse:
    def __init__(self, status: BuildStatus, payload: bytes = None, build_message: str = None, build_stderr: str = None, build_stdout: str = None):
        self.status = status
        self.payload = payload if payload is not None else b""
        self.build_message = build_message if build_message is not None else ""
        self.build_stderr = build_stderr if build_stderr is not None else ""
        self.build_stdout = build_stdout if build_stdout is not None else ""

    def get_status(self) -> BuildStatus:
        return self.status

    def set_status(self, status: BuildStatus):
        self.status = status

    def get_payload(self) -> bytes:
        return self.payload

    def set_payload(self, payload: bytes):
        self.payload = payload

    def set_build_message(self, build_message: str):
        self.build_message = build_message

    def get_build_message(self) -> str:
        return self.build_message

    def set_build_stderr(self, build_stderr: str):
        self.build_stderr = build_stderr

    def get_build_stderr(self) -> str:
        return self.build_stderr

    def get_build_stdout(self) -> str:
        return self.build_stdout

    def set_build_stdout(self, build_stdout: str):
        self.build_stdout = build_stdout


class PayloadType:

    support_browser_scripts = []

    def __init__(
        self,
        uuid: str = None,
        agent_code_path: Path = None,
        c2info: [C2ProfileParameters] = None,
        commands: CommandList = None,
        wrapped_payload: str = None,
        selected_os: str = None,
    ):
        self.commands = commands
        self.base_path = agent_code_path
        self.agent_code_path = agent_code_path / "agent_code"
        self.c2info = c2info
        self.uuid = uuid
        self.selected_os = selected_os
        self.wrapped_payload = wrapped_payload

    @property
    @abstractmethod
    def name(self):
        pass

    @property
    @abstractmethod
    def file_extension(self):
        pass

    @property
    @abstractmethod
    def author(self):
        pass

    @property
    @abstractmethod
    def supported_os(self):
        pass

    @property
    @abstractmethod
    def wrapper(self):
        pass

    @property
    @abstractmethod
    def wrapped_payloads(self):
        pass

    @property
    @abstractmethod
    def note(self):
        pass

    @property
    @abstractmethod
    def supports_dynamic_loading(self):
        pass

    @property
    @abstractmethod
    def c2_profiles(self):
        pass

    @property
    @abstractmethod
    def build_parameters(self):
        pass
        
    @property
    def translation_container(self):
        pass

    @property
    def mythic_encrypts(self):
        pass

    @abstractmethod
    async def build(self) -> BuildResponse:
        pass

    def get_parameter(self, name):
        for arg in self.build_parameters:
            if arg.name == name:
                return arg.value
        return None


    async def set_and_validate_build_parameters(self, buildinfo: dict):
        # set values for all of the key-value pairs presented to us
        for bp in self.build_parameters:
            if bp.name in buildinfo and buildinfo[bp.name] is not None:
                bp.value = buildinfo[bp.name]
            if bp.required and bp.value is None:
                raise ValueError(
                    "{} is a required parameter but has no value".format(bp.name)
                )

    def get_build_instance_values(self):
        values = {}
        for bp in self.build_parameters:
            if bp.value is not None:
                values[bp.name] = bp.value
        return values

    def to_json(self):
        return {
            "ptype": self.name,
            "file_extension": self.file_extension,
            "author": self.author,
            "supported_os": ",".join([str(x) for x in self.supported_os]),
            "wrapper": self.wrapper,
            "wrapped": self.wrapped_payloads,
            "supports_dynamic_loading": self.supports_dynamic_loading,
            "note": self.note,
            "build_parameters": [b.to_json() for b in self.build_parameters],
            "c2_profiles": self.c2_profiles,
            "support_scripts": [
                a.to_json(self.base_path) for a in self.support_browser_scripts
            ],
            "translation_container": self.translation_container,
            "mythic_encrypts": self.mythic_encrypts
        }



from aio_pika import connect_robust, IncomingMessage, Message
from aio_pika.patterns import RPC
import asyncio
import uuid
from . import MythicCommandBase
import json
import base64
import pathlib
import sys
from .config import settings

c2_connection = None
rpc = None
c2_callback_queue = None
c2_channel = None
c2_loop = None
c2_futures = {}

class RPCResponse:
    def __init__(self, resp: dict):
        self._raw_resp = resp
        if resp["status"] == "success":
            self.status = MythicCommandBase.MythicRPCStatus.Success
            self.response = resp["response"] if "response" in resp else None
            self.error = None
        else:
            self.status = MythicCommandBase.MythicRPCStatus.Error
            self.error = resp["error"]
            self.response = resp["response"] if "response" in resp else None

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, status):
        self._status = status

    @property
    def error(self):
        return self._error

    @error.setter
    def error(self, error):
        self._error = error

    @property
    def response(self):
        return self._response

    @response.setter
    def response(self, response):
        self._response = response

    def __str__(self):
        return json.dumps(self._raw_resp)


class MythicBaseRPC:
    async def connect(self):
        global c2_connection
        global rpc
        global c2_callback_queue
        global c2_channel
        global c2_loop
        if c2_connection is None:
            c2_connection = await connect_robust(
                host=settings.get("host", "127.0.0.1"),
                login=settings.get("username", "mythic_user"),
                password=settings.get("password", "mythic_password"),
                virtualhost=settings.get("virtual_host", "mythic_vhost"),       
            )
            c2_channel = await c2_connection.channel()
            c2_callback_queue = await c2_channel.declare_queue(exclusive=True)
            c2_loop = asyncio.get_event_loop()
            await c2_callback_queue.consume(self.on_response)
            try:
                rpc = await RPC.create(c2_channel)
            except Exception as e:
                print("Failed to create rpc\n" + str(e))
                sys.stdout.flush()
        return self

    def on_response(self, message: IncomingMessage):
        global c2_futures
        future = c2_futures.pop(message.correlation_id)
        future.set_result(message.body)


class MythicRPC(MythicBaseRPC):
    async def get_functions(self) -> RPCResponse:
        global rpc
        await self.connect()
        try:
            output = await rpc.proxy.get_rpc_functions()
            return RPCResponse(output)
        except Exception as e:
            print(str(sys.exc_info()[-1].tb_lineno) +str(e))
            sys.stdout.flush()
            return RPCResponse({"status": "error", "error": "Failed to find function and call it in RPC, does it exist?\n" + str(e)})

    async def execute(self, function_name: str, **func_kwargs) -> RPCResponse:
        global rpc
        await self.connect()
        try:
            func = getattr(rpc.proxy, function_name)
            if func is not None and callable(func):
                output = await func(**func_kwargs)
            else:
                output = await rpc.call(function_name, kwargs=dict(**func_kwargs))
            return RPCResponse(output)
        except Exception as e:
            print(str(sys.exc_info()[-1].tb_lineno) + " " + str(e))
            sys.stdout.flush()
            return RPCResponse({"status": "error", "error": "Failed to call function:\n" + str(e)})

    async def call_c2rpc(self, n, receiver: str = None) -> RPCResponse:
        global c2_callback_queue
        global c2_channel
        global c2_loop
        global c2_futures
        await self.connect()
        correlation_id = str(uuid.uuid4())
        future = c2_loop.create_future()

        c2_futures[correlation_id] = future
        router = "{}_rpc_queue".format(receiver)
        await c2_channel.default_exchange.publish(
            Message(
                json.dumps(n).encode(),
                content_type="application/json",
                correlation_id=correlation_id,
                reply_to=c2_callback_queue.name,
            ),
            routing_key=router,
        )
        return RPCResponse(json.loads(await future))


    async def execute_c2rpc(self, c2_profile: str, function_name: str, message: str, task_id: int) -> RPCResponse:
        try:
            resp = await self.call_c2rpc(
                {"action": function_name,
                 "message": message,
                 "task_id": task_id
                 },
                c2_profile)
            return resp

        except Exception as e:
            print(str(sys.exc_info()[-1].tb_lineno) + " " + str(e))
            sys.stdout.flush()
            return RPCResponse({"status": "error", "error": "Failed to call function:\n" + str(e)})
